#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `mlbench_core.cli` package."""

import pytest
from click.testing import CliRunner

from mlbench_core.cli import cli_group

########################## GLOBALS #################################

runner = CliRunner()

CREATE_GCLOUD_DEFAULTS = {
    "machine_type": "n1-standard-4",
    "disk_size": 50,
    "num_cpus": 4,
    "num_gpus": 0,
    "gpu_type": "nvidia-tesla-k80",
    "zone": "europe-west1-b",
    "preemptible": False,
}


########################## HELPERS #################################


def get_gcloud_cmd_line(args, option_dict=None):
    """Constructs command line for gcloud commands

    Returns command line as a list, e.g. ["3", "myrelease", "-g", "1", "--gpu-type", "nvidia-tesla-k80"].

    Args:
        args (list): command-line arguments (e.g. [3, "myrelease"])
        option_dict (dict): command-line options, as in CREATE_GCLOUD_DEFAULTS

    Returns:
        list: command line constructed from inputs
    """

    flags = {
        "machine_type": "-t",
        "disk_size": "-d",
        "num_cpus": "-c",
        "num_gpus": "-g",
        "gpu_type": "--gpu-type",
        "zone": "-z",
        "preemptible": "-e",
        "project": "-p",
        "custom_value": "-v",
    }

    args = list(map(str, args))

    def get_array(option):
        # e.g. option='num_gpus' --> ['-g', 0]
        if options[option] == True:  # check if flag is boolean
            return [flags[option]]
        elif options[option] == False:
            return []
        else:
            return [flags[option], options[option]]

    if option_dict is not None:
        options = option_dict

        cmd_line_options = [
            str(elem) for option in options for elem in get_array(option)
        ]

        return cmd_line_options + args

    return args


def create_gcloud_test_helper(mocker, gcloud_mock, args, option_dict=None):
    """Tests creation of gcloud cluster

    Tests that the correct arguments are passed to Gcloud, Tiller, etc... based
    on the input arguments and options.

    Args:
        args (list): command-line arguments
        option_dict (dict): command-line options, as in CREATE_GCLOUD_DEFAULTS. Any
                            missing options are substituted with the defaults.
    """

    container_v1 = gcloud_mock["containerv1"]
    deploy_chart = gcloud_mock["deploy_chart"]
    auth = gcloud_mock["auth"]

    cluster = container_v1.types.Cluster
    get_operation = container_v1.ClusterManagerClient.return_value.get_operation
    nodeconfig = container_v1.types.NodeConfig
    accelerator = container_v1.types.AcceleratorConfig

    if option_dict is not None and "project" in option_dict:
        project = option_dict["project"]
    else:
        _, project = auth()

    cmd_line = get_gcloud_cmd_line(args, option_dict)

    if option_dict is None:
        option_dict = CREATE_GCLOUD_DEFAULTS
    else:
        for k in CREATE_GCLOUD_DEFAULTS:  # add missing defaults
            if k not in option_dict:
                option_dict[k] = CREATE_GCLOUD_DEFAULTS[k]

    result = runner.invoke(
        cli_group, ["create-cluster", "gcloud"] + cmd_line, catch_exceptions=False
    )

    assert result.exit_code == 0

    cluster.assert_called_once()
    assert cluster.call_args[1]["initial_node_count"] == args[0]  # num_workers

    nodeconfig.assert_called_once()
    assert nodeconfig.call_args[1]["machine_type"] == option_dict["machine_type"]
    assert nodeconfig.call_args[1]["disk_size_gb"] == option_dict["disk_size"]
    assert nodeconfig.call_args[1]["preemptible"] == option_dict["preemptible"]

    deploy_chart.assert_called_once()
    assert deploy_chart.call_args[1]["num_workers"] == args[0] - 1
    assert deploy_chart.call_args[1]["num_gpus"] == option_dict["num_gpus"]
    assert deploy_chart.call_args[1]["num_cpus"] == option_dict["num_cpus"] - 1

    get_operation.assert_called()
    assert option_dict["zone"] in get_operation.call_args[1]["name"]
    assert project in get_operation.call_args[1]["name"]

    if option_dict["num_gpus"] > 0:
        num_gpus = option_dict["num_gpus"]
        gpu_type = option_dict["gpu_type"]
        accelerator.assert_called_once_with(
            accelerator_count=num_gpus, accelerator_type=gpu_type
        )


########################## FIXTURES #################################


@pytest.fixture
def gcloud_auth(mocker):
    """Patches google.auth.default to bypass authentication"""

    m = mocker.patch("google.auth.default")
    m.return_value = (m.credentials, "test_project")

    return m


@pytest.fixture
def gcloud_mock(mocker, gcloud_auth):
    """Patches all gcloud objects and makes loops run once.

    Loops like

        while response.status < response.DONE:
            #do something

    are executed once.

    Returns:
        dict: dictionary of patched objects
    """

    # One response loop
    container_v1 = mocker.patch("mlbench_core.cli.gcloud_utils.container_v1")
    gclient = container_v1.ClusterManagerClient.return_value
    response_create = gclient.create_cluster.return_value
    response_delete = gclient.delete_cluster.return_value
    response_get_op = gclient.get_operation.return_value

    response_create.status = 0
    response_create.DONE = 1
    response_create.name = "test-create-name"  # for concatenation

    response_delete.status = 0
    response_delete.DONE = 1
    response_delete.name = "test-delete-name"

    response_get_op.status = 1
    response_get_op.DONE = 1
    response_get_op.name = "test-get-op-name"

    return {
        "auth": gcloud_auth,
        "gclient": gclient,
        "containerv1": container_v1,
        "discovery": mocker.patch("mlbench_core.cli.gcloud_utils.discovery"),
        "http": mocker.patch("mlbench_core.cli.gcloud_utils.http"),
        "apiclient": mocker.patch("mlbench_core.cli.cli.ApiClient"),
        "k8sclient_gcloud": mocker.patch("mlbench_core.cli.gcloud_utils.kube_client"),
        "deploy_chart": mocker.patch("mlbench_core.cli.cli.deploy_chart"),
        "create_kubeconfig": mocker.patch(
            "mlbench_core.cli.cli.setup_gcloud_kube_client",
        ),
        "configparser": mocker.patch("configparser.ConfigParser"),
    }


@pytest.mark.parametrize(
    "args,option_dict",
    [
        ([3, "test"], None),  # default
        (  # non-default
            [3, "test"],
            {
                "machine_type": "my-custom-type",
                "disk_size": 20,
                "num_cpus": 5,
                "num_gpus": 0,
                "gpu_type": "my-nvidia-tesla-p100",
                "zone": "my-europe-west1-b",
                "preemptible": True,
                "project": "myproj",
            },
        ),
        ([10, "test"], {"num_gpus": 3}),  # with gpu
    ],
)
def test_create_gcloud_cluster(mocker, gcloud_mock, args, option_dict):
    """Tests creating a gcloud cluster.

    Uses default values, non-default values and GPU (see decorator).
    """
    create_gcloud_test_helper(mocker, gcloud_mock, args, option_dict)


def test_delete_gcloud_cluster(gcloud_mock):
    """Tests deleting a gcloud cluster"""

    gclient = gcloud_mock["gclient"]
    delete_cluster = gclient.delete_cluster
    get_operation = gclient.get_operation

    args = ["test-name"]
    opts = {"project": "test-proj", "zone": "test-zone"}

    cmd = get_gcloud_cmd_line(args, opts)

    result = runner.invoke(cli_group, ["delete-cluster", "gcloud"] + cmd)
    assert result.exit_code == 0

    delete_cluster.assert_called_once()
    get_operation.assert_called_once()

    for arg in list(opts.values()):
        assert arg in delete_cluster.call_args[1]["name"]
        assert arg in get_operation.call_args[1]["name"]

    assert args[0] in delete_cluster.call_args[1]["name"]
    assert delete_cluster.return_value.name in get_operation.call_args[1]["name"]



import torch
import torch.distributed as dist


def global_average(sum, count):
    def helper(array):
        array = get_backend_tensor(torch.Tensor(array))

        dist.all_reduce(array, op=dist.ReduceOp.SUM)
        return array[0] / array[1]

    avg = helper([sum, count])
    return avg


def get_backend_tensor(tensor):
    if dist.is_initialized() and dist.get_backend() == dist.Backend.NCCL:
        return tensor.cuda()
    return tensor



from .centralized import *
from .decentralized import *
from .optim import *

optimizers = {
    "sign_sgd": SignSGD,
    "sparsified_sgd": SparsifiedSGD,
    "centralized_sparsified_sgd": CentralizedSparsifiedSGD,
    "centralized_sgd": CentralizedSGD,
    "centralized_adam": CentralizedAdam,
    "power_sgd": PowerSGD,
    "decentralized_sgd": DecentralizedSGD,
}


def get_optimizer(optimizer, **kwargs):
    """Returns an object of the class specified with the argument `optimizer`.

    Args:
        optimizer (str): name of the optimizer
        **kwargs (dict, optional): additional optimizer-specific parameters. For the list of supported parameters
            for each optimizer, please look at its documentation.
    """
    return optimizers[optimizer](**kwargs)



import os
from collections import defaultdict
from functools import partial

import torch

from mlbench_core.dataset.nlp.pytorch.wmt16 import wmt16_config


def _pad_vocabulary(vocab, math):
    """
    Pads vocabulary to a multiple of 'pad' tokens.

    Args:
        vocab (list): list with vocabulary
        math (str): Math precision. either `fp_16`, `manual_fp16` or `fp32`

    Returns:
        list: padded vocabulary
    """
    if math == "fp16" or math == "amp_fp16":
        pad = 8
    elif math == "fp32":
        pad = 1
    else:
        raise NotImplementedError()

    vocab_size = len(vocab)
    padded_vocab_size = (vocab_size + pad - 1) // pad * pad
    for i in range(0, padded_vocab_size - vocab_size):
        token = f"madeupword{i:04d}"
        vocab.append(token)
    assert len(vocab) % pad == 0
    return vocab


class WMT16Tokenizer:
    """Tokenizer Class for WMT16 that uses the whole vocabulary

    Args:
        base_dir (str): Base directory for files
        math_precision (str): Math precision
        separator (str): BPE
    """

    def __init__(
        self,
        base_dir,
        math_precision=None,
        separator="@@",
    ):
        self.separator = separator

        vocab = [
            wmt16_config.PAD_TOKEN,
            wmt16_config.UNK_TOKEN,
            wmt16_config.BOS_TOKEN,
            wmt16_config.EOS_TOKEN,
        ]
        vocab_fname = os.path.join(base_dir, wmt16_config.VOCAB_FNAME)

        with open(vocab_fname, encoding="utf-8") as vfile:
            for line in vfile:
                vocab.append(line.strip())

        vocab = _pad_vocabulary(vocab, math_precision)
        self.vocab_size = len(vocab)

        self.tok2idx = defaultdict(partial(int, wmt16_config.UNK))
        for idx, token in enumerate(vocab):
            self.tok2idx[token] = idx

        self.idx2tok = {}
        for key, value in self.tok2idx.items():
            self.idx2tok[value] = key

    def segment(self, line):
        """
        Tokenizes single sentence and adds special BOS and EOS tokens.

        :param line: sentence

        returns: list representing tokenized sentence
        """
        line = line.strip().split()
        entry = [self.tok2idx[i] for i in line]
        entry = [wmt16_config.BOS] + entry + [wmt16_config.EOS]
        return entry

    def detokenize(self, inputs, delim=" "):
        """
        Detokenizes single sentence and removes token separator characters.

        :param inputs: sequence of tokens
        :param delim: tokenization delimiter

        returns: string representing detokenized sentence
        """
        detok = delim.join([self.idx2tok[idx] for idx in inputs])
        detok = detok.replace(self.separator + " ", "")
        detok = detok.replace(self.separator, "")

        detok = detok.replace(wmt16_config.BOS_TOKEN, "")
        detok = detok.replace(wmt16_config.EOS_TOKEN, "")
        detok = detok.replace(wmt16_config.PAD_TOKEN, "")
        detok = detok.strip()
        return detok



# Copyright (c) 2017-present, Facebook, Inc.
# All rights reserved.
#
# This source code is licensed under the license found in the LICENSE file in
# the root directory of this source tree. An additional grant of patent rights
# can be found in the PATENTS file in the same directory.

from collections import defaultdict

import torch
import torch.nn.functional as F
from torch import nn
from torch.autograd.variable import Variable
from torch.nn import Parameter

import mlbench_core.models.pytorch.transformer.modules.strided_batched_gemm as strided_batched_gemm


class QueryLinear(torch.autograd.Function):
    @staticmethod
    def forward(ctx, input, weights_q, scale):
        s = Variable(torch.tensor([scale]))
        ctx.save_for_backward(input, weights_q, s)
        q = torch.addmm(
            input.view(input.size(0) * input.size(1), input.size(2)),
            input.view(input.size(0) * input.size(1), input.size(2)),
            weights_q,
            beta=0.0,
            alpha=s[0],
        )
        q = q.view(input.size(0), input.size(1), input.size(2))
        return q.detach()

    @staticmethod
    def backward(ctx, q_grad):
        input, weights_q, s = ctx.saved_tensors
        input = input.view(input.size(0) * input.size(1), input.size(2)).transpose(0, 1)
        q = torch.addmm(
            q_grad.view(q_grad.size(0) * q_grad.size(1), q_grad.size(2)),
            q_grad.view(q_grad.size(0) * q_grad.size(1), q_grad.size(2)),
            weights_q.transpose(0, 1),
            beta=0.0,
            alpha=s[0],
        )
        q = q.view(q_grad.size(0), q_grad.size(1), q_grad.size(2))
        q_grad = q_grad.view(q_grad.size(0) * q_grad.size(1), q_grad.size(2))
        weights_q_grad = torch.addmm(weights_q, input, q_grad, beta=0.0, alpha=s[0])
        return q, weights_q_grad, None


class KeyValueLinears(torch.autograd.Function):
    @staticmethod
    def forward(ctx, input, weights_k, weights_v):
        ctx.save_for_backward(input, weights_k, weights_v)
        k = torch.addmm(
            input.view(input.size(0) * input.size(1), input.size(2)),
            input.view(input.size(0) * input.size(1), input.size(2)),
            weights_k,
            beta=0.0,
            alpha=1.0,
        )
        k = k.view(input.size(0), input.size(1), input.size(2))
        v = torch.addmm(
            input.view(input.size(0) * input.size(1), input.size(2)),
            input.view(input.size(0) * input.size(1), input.size(2)),
            weights_v,
            beta=0.0,
            alpha=1.0,
        )
        v = v.view(input.size(0), input.size(1), input.size(2))
        return k.detach(), v.detach()

    @staticmethod
    def backward(ctx, k_grad, v_grad):
        input, weights_k, weights_v = ctx.saved_tensors
        input = input.view(input.size(0) * input.size(1), input.size(2)).transpose(0, 1)
        k = torch.addmm(
            k_grad.view(k_grad.size(0) * k_grad.size(1), k_grad.size(2)),
            k_grad.view(k_grad.size(0) * k_grad.size(1), k_grad.size(2)),
            weights_k.transpose(0, 1),
            beta=0.0,
        )
        k_grad = k_grad.view(k_grad.size(0) * k_grad.size(1), k_grad.size(2))
        weights_k_grad = torch.mm(input, k_grad)
        v = k.addmm_(
            v_grad.view(v_grad.size(0) * v_grad.size(1), v_grad.size(2)),
            weights_v.transpose(0, 1),
            beta=1.0,
        )
        v = v.view(v_grad.size(0), v_grad.size(1), v_grad.size(2))
        v_grad = v_grad.view(v_grad.size(0) * v_grad.size(1), v_grad.size(2))
        weights_v_grad = torch.mm(input, v_grad)
        return v, weights_k_grad, weights_v_grad


class SelfAttentionLinears(torch.autograd.Function):
    @staticmethod
    def forward(ctx, input, weights_q, weights_k, weights_v, scale):
        s = Variable(torch.tensor([scale]))
        ctx.save_for_backward(input, weights_q, weights_k, weights_v, s)
        q = torch.addmm(
            input.view(input.size(0) * input.size(1), input.size(2)),
            input.view(input.size(0) * input.size(1), input.size(2)),
            weights_q,
            beta=0.0,
            alpha=s[0],
        )
        q = q.view(input.size(0), input.size(1), input.size(2))
        k = torch.addmm(
            input.view(input.size(0) * input.size(1), input.size(2)),
            input.view(input.size(0) * input.size(1), input.size(2)),
            weights_k,
            beta=0.0,
            alpha=1.0,
        )
        k = k.view(input.size(0), input.size(1), input.size(2))
        v = torch.addmm(
            input.view(input.size(0) * input.size(1), input.size(2)),
            input.view(input.size(0) * input.size(1), input.size(2)),
            weights_v,
            beta=0.0,
            alpha=1.0,
        )
        v = v.view(input.size(0), input.size(1), input.size(2))
        return q.detach(), k.detach(), v.detach()

    @staticmethod
    def backward(ctx, q_grad, k_grad, v_grad):
        input, weights_q, weights_k, weights_v, s = ctx.saved_tensors
        input = input.view(input.size(0) * input.size(1), input.size(2)).transpose(0, 1)
        q = torch.addmm(
            q_grad.view(q_grad.size(0) * q_grad.size(1), q_grad.size(2)),
            q_grad.view(q_grad.size(0) * q_grad.size(1), q_grad.size(2)),
            weights_q.transpose(0, 1),
            beta=0.0,
            alpha=s[0],
        )
        q_grad = q_grad.view(q_grad.size(0) * q_grad.size(1), q_grad.size(2))
        weights_q_grad = torch.addmm(weights_q, input, q_grad, beta=0.0, alpha=s[0])
        k = q.addmm_(
            k_grad.view(k_grad.size(0) * k_grad.size(1), k_grad.size(2)),
            weights_k.transpose(0, 1),
            beta=1.0,
        )
        k_grad = k_grad.view(k_grad.size(0) * k_grad.size(1), k_grad.size(2))
        weights_k_grad = torch.mm(input, k_grad)
        v = k.addmm_(
            v_grad.view(v_grad.size(0) * v_grad.size(1), v_grad.size(2)),
            weights_v.transpose(0, 1),
            beta=1.0,
        )
        v = v.view(v_grad.size(0), v_grad.size(1), v_grad.size(2))
        v_grad = v_grad.view(v_grad.size(0) * v_grad.size(1), v_grad.size(2))
        weights_v_grad = torch.mm(input, v_grad)
        return v, weights_q_grad, weights_k_grad, weights_v_grad, None


class StridedBmm1Func(torch.autograd.Function):
    @staticmethod
    def forward(ctx, input1, input2):
        ctx.save_for_backward(input1, input2)
        output = torch.empty(
            (input1.size(0), input1.size(1), input2.size(2)),
            dtype=input1.dtype,
            device=torch.device("cuda"),
        )
        if (input1.dtype == torch.float16) and (input2.dtype == torch.float16):
            output = strided_batched_gemm.strided_batched_gemm(
                0.0, output, 1.0, input1, input2
            )
        else:
            output = torch.bmm(input1, input2, out=output)
        return output.detach()

    @staticmethod
    def backward(ctx, grad_output):
        input1, input2 = ctx.saved_tensors
        grad_input1 = torch.empty(
            (input1.size(1), input2.size(0), input1.size(2)),
            dtype=input1.dtype,
            device=torch.device("cuda"),
        ).transpose(1, 0)
        grad_input2 = torch.empty(
            (input2.size(2), input2.size(0), input2.size(1)),
            dtype=input2.dtype,
            device=torch.device("cuda"),
        ).transpose(1, 0)
        if (
            (grad_output.dtype == torch.float16)
            and (input1.dtype == torch.float16)
            and (input2.dtype == torch.float16)
        ):
            grad_input1 = strided_batched_gemm.strided_batched_gemm(
                0.0, grad_input1, 1.0, grad_output, input2.transpose(1, 2)
            )
            grad_input2 = strided_batched_gemm.strided_batched_gemm(
                0.0, grad_input2, 1.0, grad_output.transpose(1, 2), input1
            )
            grad_input2 = grad_input2.transpose(1, 2)
        else:
            grad_input1 = torch.bmm(
                grad_output, input2.transpose(1, 2), out=grad_input1
            )
            grad_input2 = torch.bmm(
                grad_output.transpose(1, 2), input1, out=grad_input2
            ).transpose(1, 2)
        return grad_input1, grad_input2


class StridedBmm2Func(torch.autograd.Function):
    @staticmethod
    def forward(ctx, input1, input2):
        ctx.save_for_backward(input1, input2)
        output = torch.empty(
            (input1.size(1), input1.size(0), input2.size(2)),
            dtype=input1.dtype,
            device=torch.device("cuda"),
        ).transpose(1, 0)
        if (input1.dtype == torch.float16) and (input2.dtype == torch.float16):
            output = strided_batched_gemm.strided_batched_gemm(
                0.0, output, 1.0, input1, input2
            )
        else:
            output = torch.bmm(input1, input2, out=output)
        return output.detach()

    @staticmethod
    def backward(ctx, grad_output):
        input1, input2 = ctx.saved_tensors
        grad_input2 = torch.empty(
            (input2.size(1), input2.size(0), input2.size(2)),
            dtype=input2.dtype,
            device=torch.device("cuda"),
        ).transpose(1, 0)
        grad_input1 = torch.empty(
            (input1.size(0), input1.size(1), input1.size(2)),
            dtype=input2.dtype,
            device=torch.device("cuda"),
        )
        if (
            (grad_output.dtype == torch.float16)
            and (input1.dtype == torch.float16)
            and (input2.dtype == torch.float16)
        ):
            grad_input1 = strided_batched_gemm.strided_batched_gemm(
                0.0, grad_input1, 1.0, grad_output, input2.transpose(1, 2)
            )
            grad_input2 = strided_batched_gemm.strided_batched_gemm(
                0.0, grad_input2, 1.0, input1.transpose(1, 2), grad_output
            )
        else:
            grad_input1 = torch.bmm(grad_output, input2.transpose(1, 2))
            grad_input2 = torch.bmm(
                input1.transpose(1, 2), grad_output, out=grad_input2
            )
        return grad_input1, grad_input2


class FillPadding(torch.autograd.Function):
    @staticmethod
    def forward(ctx, inputs, mask, num_heads):
        assert len(inputs.size()) == 3, "Tensor is not 3 dims"
        bsz = int(inputs.size(0) / num_heads)
        tgt_len = inputs.size(1)
        src_len = inputs.size(2)
        inputs = inputs.view(bsz, num_heads, tgt_len, src_len)
        inputs = inputs + mask.unsqueeze(1).unsqueeze(2)
        inputs = inputs.view(bsz * num_heads, tgt_len, src_len)
        return inputs.detach()

    def backward(ctx, grad_output):
        grad_input = grad_output
        return grad_input, None, None


query_linear = QueryLinear.apply
key_value_linears = KeyValueLinears.apply
self_attn_linears = SelfAttentionLinears.apply
strided_bmm1 = StridedBmm1Func.apply
strided_bmm2 = StridedBmm2Func.apply
fill_padding_func = FillPadding.apply


class MultiheadAttention(nn.Module):
    """Multi-headed attention.

    See "Attention Is All You Need" for more details.
    """

    def __init__(
        self, embed_dim, num_heads, dropout=0.0, softmax_type="default", bias=False
    ):
        super().__init__()
        self.embed_dim = embed_dim
        self.num_heads = num_heads
        self.dropout = dropout
        self.softmax_type = softmax_type
        self.head_dim = embed_dim // num_heads
        assert (
            self.head_dim * num_heads == self.embed_dim
        ), "embed_dim must be divisible by num_heads"
        self.scaling = self.head_dim ** -0.5
        self._mask = None

        self.in_proj_weight_q = Parameter(torch.Tensor(embed_dim, embed_dim))
        self.in_proj_weight_k = Parameter(torch.Tensor(embed_dim, embed_dim))
        self.in_proj_weight_v = Parameter(torch.Tensor(embed_dim, embed_dim))

        if bias:
            self.in_proj_bias_q = Parameter(torch.Tensor(embed_dim))
            self.in_proj_bias_k = Parameter(torch.Tensor(embed_dim))
            self.in_proj_bias_v = Parameter(torch.Tensor(embed_dim))
        else:
            self.register_parameter("in_proj_bias_k", None)
            self.register_parameter("in_proj_bias_q", None)
            self.register_parameter("in_proj_bias_v", None)

        self.out_proj = nn.Linear(embed_dim, embed_dim, bias=bias)

        self.reset_parameters()

    def reset_parameters(self):
        nn.init.xavier_uniform_(self.in_proj_weight_q)
        nn.init.xavier_uniform_(self.in_proj_weight_k)
        nn.init.xavier_uniform_(self.in_proj_weight_v)
        nn.init.xavier_uniform_(self.out_proj.weight)
        if self.in_proj_bias_k is not None:
            nn.init.constant_(self.in_proj_bias_q, 0.0)
            nn.init.constant_(self.in_proj_bias_k, 0.0)
            nn.init.constant_(self.in_proj_bias_v, 0.0)
            nn.init.constant_(self.out_proj.bias, 0.0)

    def forward(
        self,
        query,
        key,
        value,
        mask_future_timesteps=False,
        key_padding_mask=None,
        incremental_state=None,
        need_weights=True,
        static_kv=False,
    ):
        """Input shape: Time x Batch x Channel

        Self-attention can be implemented by passing in the same arguments for
        query, key and value. Future timesteps can be masked with the
        `mask_future_timesteps` argument. Padding elements can be excluded from
        the key by passing a binary ByteTensor (`key_padding_mask`) with shape:
        batch x src_len, where padding elements are indicated by 1s.
        """

        qkv_same = query.data_ptr() == key.data_ptr() == value.data_ptr()
        kv_same = key.data_ptr() == value.data_ptr()

        tgt_len, bsz, embed_dim = query.size()
        assert embed_dim == self.embed_dim
        assert list(query.size()) == [tgt_len, bsz, embed_dim]
        assert key.size() == value.size()

        if incremental_state is not None:
            saved_state = self._get_input_buffer(incremental_state)
            if "prev_key" in saved_state:
                # previous time steps are cached - no need to recompute
                # key and value if they are static
                if static_kv:
                    assert kv_same and not qkv_same
                    key = value = None
        else:
            saved_state = None

        if qkv_same:
            # self-attention
            q, k, v = self_attn_linears(
                query,
                self.in_proj_weight_q,
                self.in_proj_weight_k,
                self.in_proj_weight_v,
                self.scaling,
            )
        elif kv_same:
            # encoder-decoder attention
            q = query_linear(query, self.in_proj_weight_q, self.scaling)
            if key is None:
                assert value is None
                # this will allow us to concat it with previous value and get
                # just get the previous value
                k = v = q.new(0)
            else:
                k, v = key_value_linears(
                    key, self.in_proj_weight_k, self.in_proj_weight_v
                )
        else:

            q = torch.addmm(
                query.view(query.size(0) * query.size(1), query.size(2)),
                query.view(query.size(0) * query.size(1), query.size(2)),
                self.in_proj_weight_q,
                beta=0.0,
                alpha=self.scaling,
            )
            if key is None:
                assert value is None
                # this will allow us to concat it with previous value and get
                # just get the previous value
                k = v = q.new(0)
            else:
                k = F.linear(key, self.in_proj_weight_k, self.in_proj_bias_k)
                v = F.linear(value, self.in_proj_weight_v, self.in_proj_bias_v)

        if saved_state is not None:
            if "prev_key" in saved_state:
                k = torch.cat((saved_state["prev_key"], k), dim=0)
            if "prev_value" in saved_state:
                v = torch.cat((saved_state["prev_value"], v), dim=0)
            saved_state["prev_key"] = k
            saved_state["prev_value"] = v
            self._set_input_buffer(incremental_state, saved_state)

        src_len = k.size(0)

        if key_padding_mask is not None:
            assert key_padding_mask.size(0) == bsz
            assert key_padding_mask.size(1) == src_len

        q = (
            q.contiguous()
            .view(tgt_len, bsz * self.num_heads, self.head_dim)
            .transpose(0, 1)
        )
        k = (
            k.contiguous()
            .view(src_len, bsz * self.num_heads, self.head_dim)
            .transpose(0, 1)
        )
        v = (
            v.contiguous()
            .view(src_len, bsz * self.num_heads, self.head_dim)
            .transpose(0, 1)
        )

        attn_weights = strided_bmm1(q, k.transpose(1, 2))
        assert list(attn_weights.size()) == [bsz * self.num_heads, tgt_len, src_len]

        # only apply masking at training time (when incremental state is None)
        if mask_future_timesteps and incremental_state is None:
            assert (
                query.size() == key.size()
            ), "mask_future_timesteps only applies to self-attention"
            attn_weights += self.buffered_mask(attn_weights).unsqueeze(0)

        if self.softmax_type == "fast_fill":
            if key_padding_mask is not None:
                attn_weights = fill_padding_func(
                    attn_weights, key_padding_mask, self.num_heads
                )
            attn_weights = F.softmax(attn_weights, dim=-1)
        elif self.softmax_type == "default":
            if key_padding_mask is not None:
                attn_weights = attn_weights.view(bsz, self.num_heads, tgt_len, src_len)
                attn_weights = attn_weights.masked_fill(
                    key_padding_mask.unsqueeze(1).unsqueeze(2), float("-inf")
                )
                attn_weights = attn_weights.view(bsz * self.num_heads, tgt_len, src_len)
            attn_weights = F.softmax(attn_weights, dim=-1)
        else:
            assert False, "Unknown Softmax Type: {}".format(self.softmax_type)
        attn_weights = F.dropout(attn_weights, p=self.dropout, training=self.training)
        attn = strided_bmm2(attn_weights, v)
        assert list(attn.size()) == [bsz * self.num_heads, tgt_len, self.head_dim]
        attn = attn.transpose(0, 1).contiguous().view(tgt_len, bsz, embed_dim)
        attn = self.out_proj(attn)

        if need_weights:
            # average attention weights over heads
            attn_weights = attn_weights.view(bsz, self.num_heads, tgt_len, src_len)
            attn_weights = attn_weights.sum(dim=1) / self.num_heads
        else:
            attn_weights = None

        return attn, attn_weights

    def in_proj_qkv(self, query):
        return self._in_proj(query).chunk(3, dim=-1)

    def in_proj_kv(self, key):
        return self._in_proj(key, start=self.embed_dim).chunk(2, dim=-1)

    def in_proj_q(self, query):
        return self._in_proj(query, end=self.embed_dim)

    def in_proj_k(self, key):
        return self._in_proj(key, start=self.embed_dim, end=2 * self.embed_dim)

    def in_proj_v(self, value):
        return self._in_proj(value, start=2 * self.embed_dim)

    def _in_proj(self, input, start=None, end=None):
        weight = self.in_proj_weight
        bias = self.in_proj_bias
        if end is not None:
            weight = weight[:end, :]
            if bias is not None:
                bias = bias[:end]
        if start is not None:
            weight = weight[start:, :]
            if bias is not None:
                bias = bias[start:]
        return F.linear(input, weight, bias)

    def buffered_mask(self, tensor):
        dim = tensor.size(-1)
        if self._mask is None:
            self._mask = torch.triu(fill_with_neg_inf(tensor.new(dim, dim)), 1)
        if self._mask.size(0) < dim:
            self._mask = torch.triu(fill_with_neg_inf(self._mask.resize_(dim, dim)), 1)
        return self._mask[:dim, :dim]

    def reorder_incremental_state(self, incremental_state, new_order):
        """Reorder buffered internal state (for incremental generation)."""
        input_buffer = self._get_input_buffer(incremental_state)
        if input_buffer is not None:
            for k in input_buffer.keys():
                input_buffer[k] = input_buffer[k].index_select(1, new_order)
            self._set_input_buffer(incremental_state, input_buffer)

    def _get_input_buffer(self, incremental_state):
        return (
            get_incremental_state(
                self,
                incremental_state,
                "attn_state",
            )
            or {}
        )

    def _set_input_buffer(self, incremental_state, buffer):
        set_incremental_state(
            self,
            incremental_state,
            "attn_state",
            buffer,
        )


def fill_with_neg_inf(t):
    """FP16-compatible function that fills a tensor with -inf."""
    return t.float().fill_(float("-inf")).type_as(t)


INCREMENTAL_STATE_INSTANCE_ID = defaultdict(lambda: 0)


def _get_full_incremental_state_key(module_instance, key):
    module_name = module_instance.__class__.__name__

    # assign a unique ID to each module instance, so that incremental state is
    # not shared across module instances
    if not hasattr(module_instance, "_mlbench_instance_id"):
        INCREMENTAL_STATE_INSTANCE_ID[module_name] += 1
        module_instance._mlbench_instance_id = INCREMENTAL_STATE_INSTANCE_ID[
            module_name
        ]

    return "{}.{}.{}".format(module_name, module_instance._mlbench_instance_id, key)


def get_incremental_state(module, incremental_state, key):
    """Helper for getting incremental state for an nn.Module."""
    full_key = _get_full_incremental_state_key(module, key)
    if incremental_state is None or full_key not in incremental_state:
        return None
    return incremental_state[full_key]


def set_incremental_state(module, incremental_state, key, value):
    """Helper for setting incremental state for an nn.Module."""
    if incremental_state is not None:
        full_key = _get_full_incremental_state_key(module, key)
        incremental_state[full_key] = value



import torch.nn.functional as F
from torch import nn

from .multihead_attention import MultiheadAttention


def Linear(in_features, out_features, bias=True):
    m = nn.Linear(in_features, out_features, bias)
    nn.init.xavier_uniform_(m.weight)
    nn.init.constant_(m.bias, 0.0)
    return m


def _dropout_add(x, residual, prob, is_training):
    out = F.dropout(x, p=prob, training=is_training)
    out = residual + out
    return out


def _relu_dropout(x, prob, is_training):
    out = F.threshold(x, 0.0, 0.0)
    out = F.dropout(out, p=prob, training=is_training)
    return out


class TransformerEncoderLayer(nn.Module):
    """Encoder layer block.

    In the original paper each operation (multi-head attention or FFN) is
    postprocessed with: `dropout -> add residual -> layernorm`. In the
    tensor2tensor code they suggest that learning is more robust when
    preprocessing each layer with layernorm and postprocessing with:
    `dropout -> add residual`. We default to the approach in the paper, but the
    tensor2tensor approach can be enabled by setting
    *args.encoder_normalize_before* to ``True``.

    Args:
        args (argparse.Namespace): parsed command-line arguments
    """

    def __init__(self, args):
        super().__init__()
        self.embed_dim = args.encoder_embed_dim
        self.self_attn = MultiheadAttention(
            embed_dim=self.embed_dim,
            num_heads=args.encoder_attention_heads,
            dropout=args.attention_dropout,
            bias=False,
            softmax_type=args.softmax_type,
        )
        self.dropout = args.dropout
        self.relu_dropout = args.relu_dropout
        self.normalize_before = args.encoder_normalize_before
        self.fc1 = Linear(self.embed_dim, args.encoder_ffn_embed_dim)
        self.fc2 = Linear(args.encoder_ffn_embed_dim, self.embed_dim)
        self.layer_norms = nn.ModuleList(
            [nn.LayerNorm(self.embed_dim) for _ in range(2)]
        )

    def forward(self, x, encoder_padding_mask):
        residual = x

        x = self.maybe_layer_norm(0, x, before=True)
        x, _ = self.self_attn(
            query=x, key=x, value=x, key_padding_mask=encoder_padding_mask
        )

        x = _dropout_add(x, residual, self.dropout, self.training)

        x = self.maybe_layer_norm(0, x, after=True)

        residual = x
        x = self.maybe_layer_norm(1, x, before=True)

        # Fuse Bias to GEMMs by making Tensors 2D
        sent_len, sents, hid_dim = x.size()
        x = x.view(sent_len * sents, hid_dim)
        x = _relu_dropout(self.fc1(x), self.relu_dropout, self.training)

        x = self.fc2(x)
        x = x.view(sent_len, sents, hid_dim)

        x = _dropout_add(x, residual, self.dropout, self.training)
        x = self.maybe_layer_norm(1, x, after=True)
        return x

    def maybe_layer_norm(self, i, x, before=False, after=False):
        assert before ^ after
        if after ^ self.normalize_before:
            return self.layer_norms[i](x)
        else:
            return x


class TransformerDecoderLayer(nn.Module):
    """Decoder layer block.

    In the original paper each operation (multi-head attention, encoder
    attention or FFN) is postprocessed with: `dropout -> add residual ->
    layernorm`. In the tensor2tensor code they suggest that learning is more
    robust when preprocessing each layer with layernorm and postprocessing with:
    `dropout -> add residual`. We default to the approach in the paper, but the
    tensor2tensor approach can be enabled by setting
    *args.decoder_normalize_before* to ``True``.

    Args:
        args (argparse.Namespace): parsed command-line arguments
        no_encoder_attn (bool, optional): whether to attend to encoder outputs
            (default: False).
    """

    def __init__(self, args, no_encoder_attn=False):
        super().__init__()
        self.embed_dim = args.decoder_embed_dim
        self.self_attn = MultiheadAttention(
            embed_dim=self.embed_dim,
            num_heads=args.decoder_attention_heads,
            dropout=args.attention_dropout,
            bias=False,
            softmax_type=args.softmax_type,
        )
        self.dropout = args.dropout
        self.relu_dropout = args.relu_dropout
        self.normalize_before = args.decoder_normalize_before

        self.self_attn_layer_norm = nn.LayerNorm(self.embed_dim)

        if no_encoder_attn:
            self.encoder_attn = None
            self.encoder_attn_layer_norm = None
        else:
            self.encoder_attn = MultiheadAttention(
                embed_dim=self.embed_dim,
                num_heads=args.decoder_attention_heads,
                dropout=args.attention_dropout,
                softmax_type=args.softmax_type,
                bias=False,
            )
            self.encoder_attn_layer_norm = nn.LayerNorm(self.embed_dim)

        self.fc1 = Linear(self.embed_dim, args.decoder_ffn_embed_dim)
        self.fc2 = Linear(args.decoder_ffn_embed_dim, self.embed_dim)

        self.final_layer_norm = nn.LayerNorm(self.embed_dim)
        self.need_attn = True

    def forward(self, x, encoder_out, encoder_padding_mask, incremental_state=None):
        residual = x
        x = self.maybe_layer_norm(self.self_attn_layer_norm, x, before=True)
        x, _ = self.self_attn(
            query=x,
            key=x,
            value=x,
            mask_future_timesteps=True,
            incremental_state=incremental_state,
            need_weights=False,
        )

        x = _dropout_add(x, residual, self.dropout, self.training)
        x = self.maybe_layer_norm(self.self_attn_layer_norm, x, after=True)

        attn = None
        if self.encoder_attn is not None:
            residual = x
            x = self.maybe_layer_norm(self.encoder_attn_layer_norm, x, before=True)
            x, attn = self.encoder_attn(
                query=x,
                key=encoder_out,
                value=encoder_out,
                key_padding_mask=encoder_padding_mask,
                incremental_state=incremental_state,
                static_kv=True,
                need_weights=(not self.training and self.need_attn),
            )
            x = _dropout_add(x, residual, self.dropout, self.training)
            x = self.maybe_layer_norm(self.encoder_attn_layer_norm, x, after=True)

        residual = x
        x = self.maybe_layer_norm(self.final_layer_norm, x, before=True)

        # Fuse Bias to GEMMs by making Tensors 2D
        sent_len, sents, hid_dim = x.size()
        x = x.view(sent_len * sents, hid_dim)
        x = _relu_dropout(self.fc1(x), self.relu_dropout, self.training)

        x = self.fc2(x)
        x = x.view(sent_len, sents, hid_dim)

        x = _dropout_add(x, residual, self.dropout, self.training)
        x = self.maybe_layer_norm(self.final_layer_norm, x, after=True)
        return x, attn

    def maybe_layer_norm(self, layer_norm, x, before=False, after=False):
        assert before ^ after
        if after ^ self.normalize_before:
            return layer_norm(x)
        else:
            return x

    def make_generation_fast_(self, need_attn=False, **kwargs):
        self.need_attn = need_attn



import torch
import torch.distributed as dist

from mlbench_core.aggregation.pytorch.aggregation import Aggregation


class DecentralizedAggregation(Aggregation):
    """Aggregate updates in a decentralized manner."""

    def __init__(self, rank, neighbors, use_cuda=False):
        """
        Args:
            rank (int): Rank of the current process
            neighbors (list): A list of ranks of its neighbors.
        """
        assert rank not in neighbors
        self.rank = rank
        self.neighbors = neighbors
        super(DecentralizedAggregation, self).__init__(use_cuda=use_cuda)

    def _agg(self, data, op, denom=None):
        """Aggregate data using `op` operation.

        Args:
            data (:obj:`torch.Tensor`): A Tensor to be aggragated.
            op (str): Aggregation methods like `avg`, `sum`, `min`, `max`, etc.

        Returns:
            :obj:`torch.Tensor`: An aggregated tensor.
        """
        # Create some tensors to host the values from neighborhood.
        local_data = {i: torch.zeros_like(data) for i in self.neighbors}
        local_data[self.rank] = data

        reqs = []
        for node in self.neighbors:
            reqs.append(dist.isend(tensor=local_data[self.rank], dst=node))
            reqs.append(dist.irecv(tensor=local_data[node], src=node))

        for req in reqs:
            req.wait()

        # Aggregate local_data
        if op == "avg_world":
            output = sum(local_data.values()) / (len(self.neighbors) + 1)
        else:
            raise NotImplementedError("op {} is not supported yet.".format(op))

        return output



import torch


@torch.jit.script
def orthogonalize(matrix, eps=torch.FloatTensor([1e-16])):
    """Function used to orthogonalize a matrix.

    Args:
        matrix (torch.Tensor): Matrix to orthogonalize
        eps (torch.FloatTensor): Used to avoid division by zero (default: 1e-16)
    """
    n, m = matrix.shape
    for i in range(m):
        # Normalize the i'th column
        col = matrix[:, i : i + 1]
        col /= torch.sqrt(torch.sum(col ** 2)) + eps
        # Project it on the rest and remove it
        if i + 1 < m:
            rest = matrix[:, i + 1 :]
            # rest -= torch.matmul(col.t(), rest) * col
            rest -= torch.sum(col * rest, dim=0) * col


def pack_tensors(tensors, use_cuda=False):
    """
    Packs a list of tensors into one 1-dimensional tensor.

    Args:
        tensors (list[torch.Tensor]): The tensors to pack
        use_cuda (bool): Whether the resulting tensor should be on cuda

    Returns:
        (torch.Tensor, list[int], list[(int, int)]):
            The flattened tensors, the list start indices of each packed tensor,
            and the original shape of each tensor.

            Those values are used to then unpack the tensor
    """
    indices = [0]
    for tensor in tensors:
        new_end = indices[-1] + tensor.nelement()
        indices.append(new_end)

    tensor_sizes = [t.size() for t in tensors]

    vec = torch.empty(
        indices[-1],
        device=tensors[0].device if tensors[0].is_cuda and use_cuda else "cpu",
        dtype=tensors[0].dtype,
    )

    for tensor, start_idx, end_idx in zip(tensors, indices[:-1], indices[1:]):
        vec[start_idx:end_idx] = tensor.data.view(-1)

    return vec, indices, tensor_sizes


def unpack_tensors(aggregated, indices, sizes):
    """
    Unpacks a 1-dimensional tensor into a list of tensors

    Args:
        aggregated (torch.Tensor): The 1-dimensional tensor
        indices (List[Int]): The start index of each tensor
        sizes (List[(Int, Int)]): The size of each resulting tensor

    Returns:
        List[torch.Tensor]: The unpacked tensors
    """
    start_index = indices[:-1]
    end_index = indices[1:]

    tensors = []
    for i, (start, end) in enumerate(zip(start_index, end_index)):
        tensors.append(aggregated[start:end].view(sizes[i]))

    return tensors



from mlbench_core.dataset.nlp.pytorch.wmt16.utils import build_collate_fn

from .bpttwikitext2_dataset import BPTTWikiText2
from .wmt16_dataset import WMT16Dataset
from .wmt17.batching import get_batches
from .wmt17_dataset import WMT17Dataset



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `mlbench_core.cli` package."""

import pytest
from click.testing import CliRunner

from mlbench_core.cli import cli_group

########################## GLOBALS #################################

runner = CliRunner()

########################## HELPERS #################################


def status_mock_helper(mocker, run_name):
    """Generates patched objects for "mlbench status" command.

    This is a helper function used in fixtures that test "mlbench status".

    Args:
        mocker (pytest.fixture)
        run_name (string)

    Returns:
        tuple: patched ApiClient, run ID and run name
    """
    setup = mocker.patch("mlbench_core.cli.cli.setup_client_from_config")
    client = mocker.patch("mlbench_core.cli.cli.ApiClient")
    setup.return_value = True

    rundict = {
        "name": "none" if run_name is None else run_name,
        "job_id": "",
        "job_metadata": "",
        "id": "my-id",
    }

    client.return_value.get_runs.return_value.result.return_value.json.return_value = [
        rundict
    ]

    return client, rundict["id"], rundict["name"]


########################## FIXTURES #################################


@pytest.fixture
def status_mock(mocker):
    """Patches status function when a run is found"""
    return status_mock_helper(mocker, "test-run-name")


@pytest.fixture
def status_mock_no_run(mocker):
    """Patches status function when no runs are found"""
    return status_mock_helper(mocker, None)


########################## TESTS #################################


def test_status(status_mock):
    """Tests "mlbench status" command when a run is found"""

    client, rid, name = status_mock
    url = "my/test/url"

    # skip reporting
    client.return_value.get_run_metrics.return_value.result.return_value.status_code = (
        301
    )

    get_run_metrics = client.return_value.get_run_metrics

    cmd = ["status", name, "-u", url]

    result = runner.invoke(cli_group, cmd)
    assert result.exit_code == 0

    client.assert_called_once_with(in_cluster=False, url=url, load_config=False)
    get_run_metrics.assert_called()
    assert get_run_metrics.mock_calls[0][1][0] == rid
    assert get_run_metrics.mock_calls[1][1][0] == rid


def test_status_no_run(status_mock_no_run):
    """Tests "mlbench status" command when no run is found."""

    client, rid, _ = status_mock_no_run
    get_run_metrics = client.return_value.get_run_metrics

    url = "my/test/url"
    name = "my-test-name"
    cmd = ["status", name, "-u", url]

    result = runner.invoke(cli_group, cmd)
    assert result.exit_code == 0

    client.assert_called_once_with(in_cluster=False, url=url, load_config=False)
    assert not get_run_metrics.called


def test_charts(status_mock, tmpdir):
    """Tests "mlbench charts" command on 3 finished runs"""
    folder = tmpdir.mkdir("charts")

    client, rid, name = status_mock

    # skip reporting
    client.return_value.get_run_metrics.return_value.result.return_value.status_code = (
        301
    )

    client.return_value.get_runs.return_value.result.return_value.json.return_value = [
        {"id": 1, "name": "test-0", "num_workers": 1, "state": "finished"},
        {"id": 2, "name": "test-1", "num_workers": 2, "state": "finished"},
        {"id": 3, "name": "test-2", "num_workers": 4, "state": "finished"},
    ]

    client.return_value.get_run_metrics.return_value.result.return_value.json.return_value = {
        "global_cum_agg @ 0": [{"value": "1.0"}],
        "global_cum_backprop @ 0": [{"value": "1.0"}],
        "global_cum_batch_load @ 0": [{"value": "1.0"}],
        "global_cum_comp_loss @ 0": [{"value": "1.0"}],
        "global_cum_comp_metrics @ 0": [{"value": "1.0"}],
        "global_cum_fwd_pass @ 0": [{"value": "1.0"}],
        "global_cum_opt_step @ 0": [{"value": "1.0"}],
    }

    cmd = ["charts", str(folder)]

    runner.invoke(cli_group, cmd, input="0 1 2")

    assert folder.join("total_time.png").exists()
    assert folder.join("speedup.png").exists()
    assert folder.join("time_for_all_phases.png").exists()


def test_delete_no_run(status_mock_no_run):
    """Tests mlbench delete command when no run is found"""

    client, rid, _ = status_mock_no_run
    delete_run_client = client.return_value.delete_run

    url = "my/test/url"
    name = "my-test-name"
    cmd = ["delete", name, "-u", url]

    result = runner.invoke(cli_group, cmd)
    assert result.exit_code == 0

    client.assert_called_once_with(in_cluster=False, url=url, load_config=False)
    assert not delete_run_client.called


def test_delete_run(status_mock):
    """Tests mlbench delete command when a run is found"""

    client, rid, name = status_mock
    url = "my/test/url"

    delete_run_client = client.return_value.delete_run

    cmd = ["delete", name, "-u", url]

    result = runner.invoke(cli_group, cmd)
    assert result.exit_code == 0

    client.assert_called_once_with(in_cluster=False, url=url, load_config=False)
    delete_run_client.assert_called_once_with(rid)



from .checkpoints_evaluation import CheckpointsEvaluationControlFlow
from .controlflow import (
    compute_train_batch_metrics,
    record_train_batch_stats,
    record_validation_stats,
    validation_round,
)
from .helpers import prepare_batch

__all__ = [
    "CheckpointsEvaluationControlFlow",
    "compute_train_batch_metrics",
    "record_validation_stats",
    "record_train_batch_stats",
    "validation_round",
    "prepare_batch",
]



import numpy as np
import torch

from mlbench_core.dataset.nlp.pytorch.wmt16 import wmt16_config


def build_collate_fn(sort):
    """Builds the collate function that adds the lengths of each datapoint to the batch

    Args:
        sort (bool): Sort within each batch

    Returns:
        func
    """

    def collate_seq(seq):
        """Builds batches for training or inference.
        Batches are returned as pytorch tensors, with padding.

        Args:
            seq (list[`obj`:torch.Tensor]): The batch

        Returns:
            (`obj`:torch.Tensor, `obj`:torch.Tensor): The batch and the lengths of each sequence
        """
        lengths = torch.tensor([len(s) for s in seq], dtype=torch.int64)
        batch_length = max(lengths)

        shape = (len(seq), batch_length)
        seq_tensor = torch.full(shape, wmt16_config.PAD, dtype=torch.int64)

        for i, s in enumerate(seq):
            end_seq = lengths[i]
            seq_tensor[i, :end_seq].copy_(s[:end_seq])

        seq_tensor = seq_tensor.t()

        return seq_tensor, lengths

    def parallel_collate(seqs):
        """Builds batches from parallel dataset (src, tgt), optionally sorts batch
        by src sequence length.

        Args:
            seqs (tuple): Tuple of (data, target) sequences

        Returns:
            (tuple, tuple): The data and target, along with the lengths
        """
        src_seqs, trg_seqs = zip(*seqs)
        if sort:
            indices, src_seqs = zip(
                *sorted(enumerate(src_seqs), key=lambda x: len(x[1]), reverse=True)
            )
            trg_seqs = [trg_seqs[idx] for idx in indices]

        src_seqs = collate_seq(src_seqs)
        trg_seqs = collate_seq(trg_seqs)
        return src_seqs, trg_seqs

    return parallel_collate


def get_data_dtype(vocab_size):
    """Returns the appropriate dtype for the given vocab size

    Args:
        vocab_size (int):

    Returns:
        (:obj:`np.dtype`, :obj:`torch.dtype`): One of `np.int16`, `np.int32` or `np.int64`
    """
    if vocab_size <= np.iinfo(np.int16).max:
        dtype = np.int16
        torch_dtype = torch.int16
    elif vocab_size <= np.iinfo(np.int32).max:
        dtype = np.int32
        torch_dtype = torch.int32
    elif vocab_size <= np.iinfo(np.int64).max:
        dtype = np.int64
        torch_dtype = torch.int64
    else:
        raise ValueError("Vocabulary size is too large")
    return dtype, torch_dtype


def process_raw_data(file_name, max_size=None):
    """Reads the lines of the given filename without any processing

    Args:
        file_name (str): Full file path
        max_size (int): Maximum size of result

    Returns:
        (list[str]): The read data
    """
    with open(file_name, mode="r", encoding="utf-8") as f:
        data = f.readlines()

    if max_size:
        data = data[:max_size]
    return data


def filter_raw_data(raw_src, raw_trg, min_len=0, max_len=float("inf")):
    """Filters raw data on source and target lengths

    Args:
        raw_src (list[str]): Raw source data
        raw_trg (list[str]): Raw target data
        min_len (int): Minimum length
        max_len (int): Maximum length

    Returns:
        (list[str], list[int],list[str], list[int]): The filtered data and the lengths
    """
    filtered_src = []
    filtered_trg = []
    filtered_src_len = []
    filtered_trg_len = []
    for src, trg in zip(raw_src, raw_trg):
        src_len = src.count(" ") + 1
        trg_len = trg.count(" ") + 1
        if min_len <= src_len <= max_len and min_len <= trg_len <= max_len:
            filtered_src.append(src)
            filtered_trg.append(trg)
            filtered_src_len.append(src_len)
            filtered_trg_len.append(trg_len)
    return filtered_src, filtered_src_len, filtered_trg, filtered_trg_len


def process_data(file_name, tokenizer, max_size=None):
    """Reads the line of a file and segments the lines using the tokenizer

    Args:
        file_name (str): Full file path
        tokenizer (:obj:`WMT16Tokenizer`): The tokenizer to use
        max_size (Optional[int]): The maximum size of result

    Returns:
        (list[int]): Tokenized data
    """
    data = []
    with open(file_name, mode="r", encoding="utf-8") as f:
        for idx, line in enumerate(f):
            if max_size and idx == max_size:
                break
            entry = tokenizer.segment(line)
            entry = torch.tensor(entry)
            data.append(entry)
    return data


def filter_data(src, trg, min_len=0, max_len=float("inf")):
    """Filters data on source and target lengths

    Args:
        src (list[int]): Tokenized source data
        trg (list[int]): Tokenized target data
        min_len (int): Minimum length
        max_len (int): Maximum length

    Returns:
        (list[int], list[int]): The filtered data (tokenized)
    """
    filtered_src = []
    filtered_trg = []
    for src, trg in zip(src, trg):
        if min_len <= len(src) <= max_len and min_len <= len(trg) <= max_len:
            filtered_src.append(src)
            filtered_trg.append(trg)
    return filtered_src, filtered_trg



import os
import subprocess
import tempfile

from git import Repo
from supermutes.dot import dotify

DEFAULT_GIT_BRANCH = "develop"


def git_clone(repo_url, branch="master", path=""):
    """clones repo to a temporary dir, the path of which is determined by the platform"""

    _tmp_dir = tempfile.mkdtemp(prefix="mlbench-")
    repo = Repo.clone_from(repo_url, _tmp_dir, branch=branch)

    return os.path.join(_tmp_dir, path)


class ChartBuilder:
    """Class that allows building helm charts either from a repository or a local folder

    Args:
        chart (dict): Dictionary describing the location. Should look like:
            ```
            {
             "name": [chart_name],
             "source": {
                        "type": ["git" or "directory"],
                        "location": [repo_url or directory path]
                        "reference": [optional, to select the branch]
                        }
            }
            ```
    """

    def __init__(self, chart):
        self.chart = dotify(chart)
        self.source_directory = self.source_clone()

    def source_clone(self):
        """
        Clone the charts source
        We only support a git source type right now, which can also
        handle git:// local paths as well
        """

        subpath = self.chart.source.get("subpath", "")

        if "name" not in self.chart:
            raise ValueError("Please specify name for the chart")

        if "type" not in self.chart.source:
            raise ValueError("Need source type for chart {}".format(self.chart.name))

        if self.chart.source.type == "git":
            if "reference" not in self.chart.source:
                self.chart.source.reference = DEFAULT_GIT_BRANCH
            if "path" not in self.chart.source:
                self.chart.source.path = ""
            self._source_tmp_dir = git_clone(
                self.chart.source.location,
                self.chart.source.reference,
                self.chart.source.path,
            )
        elif self.chart.source.type == "directory":
            self._source_tmp_dir = self.chart.source.location

        else:
            raise ValueError(
                "Unknown source type %s for chart %s",
                self.chart.name,
                self.chart.source.type,
            )

        return os.path.join(self._source_tmp_dir, subpath)

    def _get_values_string(self, vals, parent=None):
        """Given a dictionary of values, recursively returns the arguments to pass to `helm template`.

        For example: {"key1": "value1", "key2": {"key3":"value3"}}
            gives ["--set", "key1=value1", "--set", "key2.key3=value3"]

        Args:
            vals (dict): Dictionary of values
            parent (str, optional): The parent key

        Returns:
            (list[str]): The command list
        """
        values = []
        for k, v in vals.items():
            if type(v) == dict:
                values += self._get_values_string(v, k)
            else:
                key = "{}={}".format(k, v)
                if parent is not None:
                    key = "{}.{}".format(parent, key)

                values += ["--set", key]
        return values

    def get_chart(self, release_name, values):
        """Executes the command `helm template {args}` to generate the chart
        and saves the yaml to a temporary directory

        Args:
            release_name (str): Release name
            values (dict): Values to overwrite

        Returns:
            (str): Path of generated template
        """
        values_options = self._get_values_string(values)
        output = subprocess.check_output(
            ["helm", "template", release_name, self.source_directory] + values_options
        )

        if self.chart.source.type == "git":
            subpath = self.chart.source.get("subpath", "")
            template_path = os.path.join(
                self._source_tmp_dir, subpath, "mlbench_template.yaml"
            )
        else:
            template_path = os.path.join(tempfile.mkdtemp(), "template.yaml")

        with open(template_path, "wb") as f:
            f.write(output)
        return template_path



import math

from torch.utils.data import Dataset
from torchtext.experimental.datasets import WikiText2


class BPTTWikiText2(Dataset):
    """WikiText2 dataset with backpropagation through time.

    Args:
        bptt_len (int): Length of BPTT segments
        train (bool): Whether to get the train or validation set (default=True)
        tokenizer (:obj:`torchtext.data.utils.tokenizer`): Tokenizer to use
        root (str): Root folder for the dataset
    """

    def __init__(self, bptt_len, train=True, **kwargs):
        super(BPTTWikiText2, self).__init__()

        self.bptt_len = bptt_len

        train_set, _, val_set = WikiText2(**kwargs)
        self.vocab = train_set.get_vocab()

        if train:
            self.data = train_set
        else:
            if "vocab" in kwargs:
                del kwargs["vocab"]

            (val_set,) = WikiText2(vocab=self.vocab, data_select="valid", **kwargs)
            self.data = val_set

    def __getitem__(self, i):
        i = i * self.bptt_len
        seq_len = min(self.bptt_len, len(self.data) - i - 1)
        return self.data[i : i + seq_len], self.data[i + 1 : i + 1 + seq_len]

    def __len__(self):
        return math.ceil((len(self.data) - 1) / self.bptt_len)

    def __iter__(self):
        for i in range(0, len(self) * self.bptt_len, self.bptt_len):
            seq_len = min(self.bptt_len, len(self.data) - i - 1)
            text = self.data[i : i + seq_len]
            target = self.data[i + 1 : i + 1 + seq_len]
            yield text, target

    def get_vocab(self):
        return self.vocab



r"""Helper functions."""

import logging
import os
import random
import shutil
import socket

import numpy as np
import torch
from torch import distributed as dist

from mlbench_core.utils.pytorch.topology import FCGraph


def config_logging(logging_level="INFO", logging_file="/mlbench.log"):
    """Setup logging modules.
    A stream handler and file handler are added to default logger `mlbench`.

    Args:
        logging_level (str): Log level
        logging_file (str): Log file

    """

    class RankFilter(logging.Filter):
        def filter(self, record):
            record.rank = dist.get_rank()
            return True

    logger = logging.getLogger("mlbench")
    if len(logger.handlers) >= 2:
        return

    logger.setLevel(logging_level)
    logger.addFilter(RankFilter())

    formatter = logging.Formatter(
        "%(asctime)s %(name)s %(rank)2s %(levelname)s: %(message)s", "%Y-%m-%d %H:%M:%S"
    )

    ch = logging.StreamHandler()
    ch.setLevel(logging_level)
    ch.setFormatter(formatter)
    logger.addHandler(ch)

    fh = logging.FileHandler(logging_file)
    fh.setLevel(logging_level)
    fh.setFormatter(formatter)
    logger.addHandler(fh)


def config_pytorch(use_cuda=False, seed=None, cudnn_deterministic=False):
    """Config pytorch packages.

    Fix random number for packages and initialize distributed environment for pytorch.
    Setup cuda environment for pytorch.

    Args:
        use_cuda (bool): Use CUDA acceleration
        seed (int | None): Random seed to use
        cudnn_deterministic (bool): Set `cudnn.determenistic=True`

    Returns:
        (int, int, `obj`:FCGraph): The rank, world size, and network graph
    """
    # Setting `cudnn.deterministic = True` will turn on
    # CUDNN deterministic setting which can slow down training considerably.
    # Unexpected behavior may also be observed from checkpoint.
    # See: https: // github.com/pytorch/examples/blob/master/imagenet/main.py
    if cudnn_deterministic:
        # cudnn.deterministic = True
        print(
            "You have chosen to seed training. "
            "This will turn on the CUDNN deterministic setting, "
            "which can slow down your training considerably! "
            "You may see unexpected behavior when restarting "
            "from checkpoints."
        )

    if seed:
        torch.manual_seed(seed)
        torch.cuda.manual_seed_all(seed)
        np.random.seed(seed)
        random.seed(seed)
        os.environ["PYTHONHASHSEED"] = str(seed)

    # define the graph for the computation.
    if use_cuda:
        assert torch.cuda.is_available()

    rank = dist.get_rank()
    world_size = dist.get_world_size()
    backend = dist.get_backend() if dist.is_initialized() else None
    graph = FCGraph(rank, world_size, use_cuda)

    # enable cudnn accelerator if we are using cuda.
    if use_cuda:
        graph.assigned_gpu_id()
        torch.backends.cudnn.enabled = True
        torch.backends.cudnn.benchmark = False

        if cudnn_deterministic:
            torch.backends.cudnn.deterministic = True

        if torch.backends.cudnn.version() is None:
            print("CUDNN not found on device.")

        print(
            "World size={}, Rank={}, hostname={}, backend={}, cuda_available={}, cuda_device={}".format(
                world_size,
                rank,
                socket.gethostname(),
                backend,
                torch.cuda.is_available(),
                torch.cuda.current_device(),
            )
        )

    return rank, world_size, graph


def config_path(ckpt_run_dir, delete_existing_ckpts=False):
    """Config the path used during the experiments."""
    if delete_existing_ckpts:
        print("Remove previous checkpoint directory : {}".format(ckpt_run_dir))
        shutil.rmtree(ckpt_run_dir, ignore_errors=True)
    os.makedirs(ckpt_run_dir, exist_ok=True)



from mlbench_core.utils.pytorch.utils import pack_tensors, unpack_tensors


class Aggregation(object):
    """Aggregate updates / models from different processes.

    Args:
        use_cuda (bool): Whether to use CUDA tensors for communication
    """

    def __init__(self, use_cuda=False):
        self.use_cuda = use_cuda

    def _agg(self, data, op, denom=None):
        """Aggregate data using `op` operation.

        Args:
            data (:obj:`torch.Tensor`): A Tensor to be aggregated.
            op (str): Aggregation methods like `avg`, `sum`, `min`, `max`, etc.
            denom (:obj:`torch.Tensor`, optional): Custom denominator to average by
                Use with op == `custom_avg`. (default: `None`)

        Returns:
            :obj:`torch.Tensor`: An aggregated tensor.
        """
        raise NotImplementedError

    def _agg_weights_by_model(self, model, op, denom=None):
        """Aggregate models by model weight, all layers at once

        Args:
            model (:obj:`torch.Module`): Models to be averaged.
            op (str): Aggregation method. Should be in `ALLREDUCE_AGGREGATION_OPS`
            denom (:obj:`torch.Tensor`, optional): Custom denominator to average by
                Use with op == `custom_avg`. (default: `None`)
        """
        # Pack all layers
        packed, indices, sizes = pack_tensors(
            [t for t in model.parameters()], use_cuda=self.use_cuda
        )
        aggregated = self._agg(packed, op=op, denom=denom)

        tensors = unpack_tensors(aggregated, indices, sizes)
        # Unpack
        for i, param in enumerate(model.parameters()):
            param.data = tensors[i]

    def _agg_gradients_by_model(self, model, op, denom=None):
        """Aggregate models gradients, all layers at once

        Args:
            model (:obj:`torch.Module`): Models to be averaged.
            op (str): Aggregation method. Should be in `ALLREDUCE_AGGREGATION_OPS`
            denom (:obj:`torch.Tensor`, optional): Custom denominator to average by
                Use with op == `custom_avg`. (default: `None`)
        """
        # Pack all layers
        packed, indices, sizes = pack_tensors(
            [t.grad for t in model.parameters()], use_cuda=self.use_cuda
        )
        aggregated = self._agg(packed, op=op, denom=denom)

        # Unpack
        tensors = unpack_tensors(aggregated, indices, sizes)
        for i, param in enumerate(model.parameters()):
            param.grad.data = tensors[i]

    def _agg_weights_by_layer(self, model, op, denom=None):
        """Aggregate models by model weight, for each layer individually

        Args:
            model (:obj:`torch.Module`): Models to be averaged.
            op (str): Aggregation method. Should be in `ALLREDUCE_AGGREGATION_OPS`
            denom (:obj:`torch.Tensor`, optional): Custom denominator to average by
                Use with op == `custom_avg`. (default: `None`)
        """
        # Aggregate layer by layer
        for _, param in enumerate(model.parameters()):
            grad = self._agg(param.data, op=op, denom=denom)
            param.data = grad

    def _agg_gradients_by_layer(self, model, op, denom=None):
        """Aggregate models gradients each layer individually

        Args:
            model (:obj:`torch.Module`): Models to be averaged.
            op (str): Aggregation method. Should be in `ALLREDUCE_AGGREGATION_OPS`
            denom (:obj:`torch.Tensor`, optional): Custom denominator to average by
                Use with op == `custom_avg`. (default: `None`)
        """
        # Aggregate layer by layer
        for _, param in enumerate(model.parameters()):
            grad = self._agg(param.grad.data, op=op, denom=denom)
            param.grad.data = grad

    def agg_model(self, by_layer=False):
        if by_layer:
            return self._agg_weights_by_layer
        else:
            return self._agg_weights_by_model

    def agg_grad(self, by_layer=False):
        if by_layer:
            return self._agg_gradients_by_layer
        else:
            return self._agg_gradients_by_model



import subprocess
from time import sleep

from kubernetes import client

from mlbench_core.cli.chartbuilder import ChartBuilder


def _get_master_pod(release_name, pods):
    """Given a release name and a list of pods, returns the master pod of the release

    Args:
        release_name (str): Release name
        pods (:obj:`V1PodList`): List of pods

    Returns:
        (:obj:`Pod`, optional): The master pod
    """

    master_pod_name = "{}-mlbench-master-".format(release_name)
    for pod in pods.items:
        if master_pod_name in pod.metadata.name:
            return pod

    return None


def _wait_for_deployment(release_name):
    """Given a release name, waits for the master pod to be running

    Args:
        release_name (str): Release name

    Raises:
        ValueError: If the master pod is not running
    """
    kube_api = client.CoreV1Api()
    pods = kube_api.list_namespaced_pod(namespace="default")
    master_pod = _get_master_pod(release_name, pods)
    while master_pod is None or master_pod.status.phase == "Pending":
        pods = kube_api.list_namespaced_pod(namespace="default")
        master_pod = _get_master_pod(release_name, pods)
        sleep(1)
    if master_pod is None or master_pod.status.phase in ["Failed", "Unknown"]:
        raise ValueError("Could not deploy chart")


def deploy_chart(
    num_workers,
    num_gpus,
    num_cpus,
    release_name,
    custom_value,
    kube_context,
    custom_chart=None,
):
    """Deploys the mlbench-helm chart given its values

    Args:
        num_workers (int): Number of worker nodes (excluding master)
        num_gpus (int): Number of GPUs per node
        num_cpus (int): Number of CPUs per node
        release_name (str): Release name
        custom_value (str): Custom values for chart
        kube_context (str): Current kube-context (must be saved in kubeconfig)
        custom_chart (dict, optional): Custom chart to use (e.g. local chart)
    """
    sleep(5)

    # install chart
    chart = ChartBuilder(
        {
            "name": "mlbench-helm",
            "source": {
                "type": "git",
                "location": "https://github.com/mlbench/mlbench-helm",
            },
        }
        if custom_chart is None
        else custom_chart
    )

    values = {"limits": {"workers": num_workers, "gpu": num_gpus, "cpu": num_cpus}}
    if custom_value:
        # merge custom values with values
        for cv in custom_value:
            key, v = cv.split("=", 1)

            current = values
            key_path = key.split(".")

            for k in key_path[:-1]:
                if k not in current:
                    current[k] = {}

                current = current[k]

            current[key_path[-1]] = v

    chart_path = chart.get_chart(release_name, values)

    output = subprocess.check_output(
        [
            "kubectl",
            "apply",
            "--validate=false",
            "--context={}".format(kube_context),
            "-f",
            chart_path,
        ]
    )
    sleep(1)

    _wait_for_deployment(release_name)



r"""Control flow for pytorch applications."""
import logging

import torch

from mlbench_core.controlflow.pytorch.helpers import iterate_dataloader
from mlbench_core.utils import AverageMeter
from mlbench_core.utils.pytorch.distributed import global_average

logger = logging.getLogger("mlbench")

LOG_EVERY_N_BATCHES = 25


def compute_train_batch_metrics(loss, output, target, metrics):
    """Computes the given metrics on the given batch

    Args:
        loss (float): The loss of the batch
        output (:obj:`torch.Tensor`): The model output
        target (:obj:`torch.Tensor`): The labels for the current batch
        metrics (list): List of metrics to track

    Returns:
        (dict of :obj:`mlbench_core.evaluation.pytorch.metrics.MLBenchMetric`: float): The metric
            and its computed value
    """
    # Compute metrics for one batch
    result = {}
    for metric in metrics:
        metric_value = metric(loss, output, target).item()
        result[metric] = metric_value
    return result


def record_train_batch_stats(
    batch_idx, loss, output, metric_results, tracker, num_batches_per_device_train
):
    """Record the stats in a training batch.

    Args:
        batch_idx (int): The id of the current batch
        loss (float): The loss of the batch
        output (:obj:`torch.Tensor`): The model output
        metric_results (dict): of :obj:`mlbench_core.evaluation.pytorch.metrics.MLBenchMetric`: float Metrics and their values
        tracker (:obj:`mlbench_core.utils.Tracker`): Tracker object to use.
        num_batches_per_device_train (int): Number of batches per train epoch
    """
    progress = batch_idx / num_batches_per_device_train
    progress += tracker.current_epoch

    log_to_api = (
        batch_idx % LOG_EVERY_N_BATCHES == 0
        or batch_idx == num_batches_per_device_train
    )

    if tracker:
        tracker.record_loss(loss, output.size()[0], log_to_api=log_to_api)

        for metric, metric_value in metric_results.items():
            tracker.record_metric(
                metric, metric_value, output.size()[0], log_to_api=log_to_api
            )
    status = "Epoch {:5.2f} Batch {:4}: ".format(progress, batch_idx)

    logger.info(status + str(tracker))


def validation_round(
    dataloader,
    model,
    loss_function,
    metrics,
    dtype,
    tracker=None,
    transform_target_type=False,
    use_cuda=False,
    max_batches=None,
    init_hidden=None,
    package_hidden=None,
    transform_parameters=None,
):
    """Evaluate the model on the test dataset.

    Args:
        dataloader (`obj`:torch.utils.data.DataLoader): The validation set
        model (`obj`:torch.nn.Module): The model to train
        loss_function (`obj`:torch.nn.Module): The loss function
        metrics (list): List of metrics to track
        dtype (str): The datatype to use, one of `fp32`or `fp64`
        tracker (`obj`:mlbench_core.utils.Tracker | None): Tracker object to use.
        transform_target_type (bool): Convert target to `dtype`. Default `False`
        use_cuda (bool): Whether to use GPU for training, default: `False`
        max_batches (int | None): Maximum number of batches to validate on
        init_hidden (`func`): Function to initialize hidden state (for RNNs), default: `None`
        package_hidden (`func`): Function to (re-)package hidden state (for RNNs), default: `None`

    Returns:
          (dict, float): Dictionary of average of each metric, and average validation loss
    """

    model.eval()
    if tracker:
        tracker.validation()
        tracker.validation_start()

    # Initialize the accumulators for loss and metrics
    losses = AverageMeter()
    for metric in metrics:
        metric.reset()

    # Each worker computer their own losses and metrics
    with torch.no_grad():

        data_iter = iterate_dataloader(
            dataloader, dtype, max_batches, use_cuda, transform_target_type
        )

        hidden = None
        if init_hidden:
            hidden = init_hidden()

        for data, target in data_iter:
            if hidden:
                hidden = package_hidden(hidden)

            # Inference
            if hidden:
                output, hidden = model(data, hidden)
                target = target.view(-1)
            else:
                output = model(data)

            # Compute loss
            loss = loss_function(output, target)

            # Update loss
            losses.update(loss.item(), data.size(0))

            # Update metrics
            for metric in metrics:
                metric_value = metric(loss, output, target)
                metric.update(metric_value, data.size(0))

    # Aggregate metrics and loss for all workers
    metrics_averages = {metric: metric.average().item() for metric in metrics}
    loss_average = global_average(losses.sum, losses.count).item()

    if tracker:
        tracker.validation_end()
    return metrics_averages, loss_average


def record_validation_stats(metrics_values, loss, tracker=None, rank=0):
    """Records the stats of a previously run validation

    Args:
        metrics_values (dict): Dictionary of each metric's average.
        loss (float): Validation loss
        tracker (`obj`:mlbench_core.utils.Tracker, optional): Tracker object to use.
        rank (int): Current distributed rank

    Returns:
        (bool): Whether this validation round is the best
    """
    if len(metrics_values) > 0:
        # Save
        if tracker:
            for metric, value in metrics_values.items():
                tracker.record_metric(metric, value, log_to_api=rank == 0)

                tracker.record_stat(
                    "global_{}".format(metric.name),
                    value,
                    log_to_api=rank == 0,
                )

        if rank == 0 and tracker:
            logger.info(
                "{} for rank {}:(best epoch {}, current epoch {}): {:.3f}".format(
                    tracker.primary_metric.name,
                    tracker.rank,
                    tracker.best_epoch,
                    tracker.current_epoch,
                    tracker.best_metric_value,
                )
            )
    else:
        if rank == 0:
            logger.info("Validation loss={:.3f}".format(loss))

    if tracker:
        tracker.record_loss(loss, log_to_api=True)

        global_loss = global_average(loss, 1).item()

        if rank == 0:
            tracker.record_stat("global_loss", global_loss, log_to_api=True)

    return tracker.is_best() if tracker else False
